/*
	--- Version 2.0 89-12-13 17:45 ---

   TSKCONF.H - CTask - Configuration definitions 

   Public Domain Software written by
      Thomas Wagner
      Patschkauer Weg 31
      D-1000 Berlin 33
      West Germany

   This file is new with Version 1.1
*/

/*
   TSK_DYNAMIC    if 1, creating dynamic control blocks is allowed.

   CAUTION:    Changing this define requires changing the corresponding
               definition in TSK.MAC and recompilation/assembly of *all*
               sources. 
*/

#define  TSK_DYNAMIC    1


/*
   IBM            if 1, routines specific to the IBM family of PCs
                  (and clones) are installed.

   DOS            if 1, routines specific to PC/MS-DOS are installed.

      NOTE: 		If you disable IBM, you have to provide timer and keyboard
            		hardware support yourself. Check the areas in "tskmain.c" 
            		that are affected by this define.
						If you disable DOS, the resulting version can not be
						used under DOS without significant changes.

	   CAUTION:  	Changing this define requires changing the corresponding
               	definition in TSK.MAC and recompilation/assembly of *all*
               	sources.
*/

#define  IBM   1
#define  DOS   1


/*
   TSK_NAMEPAR    if 1, the create_xxx routines accept a name parameter.

   TSK_NAMED      if 1, control block names are processed.

	               TSK_NAMED may only be enabled (1) if TSK_NAMEPAR is also 
   	            enabled.

   GROUPS         if 1, multiple task groups are supported.
                  For this option to work, TSK_NAMEPAR and TSK_NAMED must
                  be enabled.

	   CAUTION:  	Changing this define requires changing the corresponding
               	definition in TSK.MAC and recompilation/assembly of *all*
               	sources.

	If DOS is nonzero, TSK_NAMEPAR, TSK_NAMED, and GROUPS must also be set.
*/


#if (!DOS)

#define  TSK_NAMEPAR    1     /* This may be set to 0 if desired */

#if (TSK_NAMEPAR)

#define  TSK_NAMED      1     /* This may be set to 0 if desired */

#else
#define  TSK_NAMED      0     /* Do not change this */
#endif

#if (TSK_NAMED && TSK_NAMEPAR)
#define  GROUPS         1     /* This may be set to 0 if desired */
#else
#define  GROUPS         1     /* Do not change this */
#endif

#else
#define  TSK_NAMEPAR    1     /* Do not change this */
#define  TSK_NAMED      1     /* Do not change this */
#define  GROUPS         1     /* Do not change this */
#endif


/*
   CLOCK_MSEC     if 1, all timeouts are specified in milliseconds.
                  Requires floating point library.
                  Changing requires recompile of tskmain.c and tsksub.c.
*/

#define CLOCK_MSEC      0

/*
   PRI_TIMER      Priority of the timer task. Normally higher than any
                  other task in the system.
                  Changing requires recompile of tskmain.c.
*/

#define  PRI_TIMER   0xf000

/*
   PRI_STD        standard priority of user tasks (arbitrary), for use
                  in user programs.
                  Changing requires recompile of tskmain.c.
*/

#define  PRI_STD     100

/*
   PRI_INT8       priority of the int8 task. Normally lower than or equal to
                  user defined tasks, but may be adjusted to tune responses
                  of TSR's. If there are busy waiting tasks in the system,
						the priority should be equal to or higher than that of the
						task that does the busy waiting.

                  Changing requires recompile of tskmain.c.
*/

#define  PRI_INT8    (PRI_STD)

/*
   AT_BIOS        if 1, the AT BIOS wait/post handler is installed.
						This can normally be left on even for XT type machines.
                  Changing requires recompile of tskmain.c.
*/

#define  AT_BIOS  1

/*
   SINGLE_DATA    if 1, only a single global data block is used.
                  This speeds up processing, but prohibits linkage
                  of multiple groups. Useful for dedicated systems,
                  and also if your program is known to never interact
                  with another copy of CTask. Should be left 0 if you're
                  not so sure, must be 0 if GROUPS is enabled.
*/

#if (!GROUPS)
#define  SINGLE_DATA  0          /* May be set to 1 if desired */
#else
#define  SINGLE_DATA  0          /* Do not change this */
#endif

